"""
    This module is a wrapper of the ``_metis`` extension generated
    from ``_metis.pyx``. This is to let sphinx import the module for
    documentation builds with or without the ``_metis`` extension.
"""
try:
    from nxmetis import _metis
except ImportError:
    def node_nd(*args, **kwargs):
        raise Exception("NetworkX-METIS not installed!")

    def part_graph(*args, **kwargs):
        raise Exception("NetworkX-METIS not installed!")

    def compute_vertex_separator(*args, **kwargs):
        raise Exception("NetworkX-METIS not installed!")

    def set_default_options(*args, **kwargs):
        raise Exception("NetworkX-METIS not installed!")
else:
    node_nd = _metis.node_nd
    part_graph = _metis.part_graph
    compute_vertex_separator = _metis.compute_vertex_separator
    set_default_options = _metis.set_default_options



import enum

__all__ = ['MetisPType', 'MetisObjType', 'MetisCType', 'MetisIPType',
           'MetisRType', 'MetisNumbering', 'MetisDbgLvl', 'MetisRStatus']


@enum.unique
class MetisPType(enum.IntEnum):
    """Partitioning method."""

    default = -1
    """Default partitioning method."""

    rb = 0
    """Multilevel recursive bisectioning."""

    kway = 1
    """Multilevel `k`-way partitioning."""


@enum.unique
class MetisObjType(enum.IntEnum):
    """Type of objective."""

    default = -1
    """Default type of objective."""

    cut = 0
    """Edge-cut minimization."""

    vol = 1
    """Total communication volume minimization."""


@enum.unique
class MetisCType(enum.IntEnum):
    """Catching scheme to be used during coarsening."""

    default = -1
    """Default catching scheme."""

    rm = 0
    """Random matching."""

    shem = 1
    """Sorted heavy-edge matching."""


@enum.unique
class MetisIPType(enum.IntEnum):
    """Algorithm used during initial partitioning."""

    default = -1
    """Default method for initial partitioning."""

    grow = 0
    """Grow a bisection using a greedy strategy."""
    
    random = 1
    """Compute a bisection at random followed by a refinement."""
    
    edge = 2
    """Derive a separator from an edge cut."""
    
    node = 3
    """Grow a bisection using a greedy node-based strategy."""


@enum.unique
class MetisRType(enum.IntEnum):
    """Algorithm used for refinement."""

    default = -1
    """Default method used for refinement."""

    fm = 0
    """FM-based cut refinement."""
    
    greedy = 1
    """Greedy-based cut and volume refinement."""
    
    sep2sided = 2
    """Two-sided node FM refinement."""
    
    sep1sided = 3
    """One-sided node FM refinement."""


@enum.unique
class MetisNumbering(enum.IntEnum):
    """Numbering scheme is used for the adjacency structure of a graph or the
    element-node structure of a mesh."""

    default = -1
    """Default numbering scheme."""

    zero = 0
    """C-style zero-based numbering."""

    one = 1
    """Fortran-style one-based numbering."""


@enum.unique
class MetisDbgLvl(enum.IntEnum):
    """Amount of progress/debugging information will be printed during the
    execution of the algorithms. Can be combined by bit-wise OR."""

    default = -1
    """Display default statistics."""

    info = 1
    """Print various diagnostic messages."""

    time = 2
    """Perform timing analysis."""

    coarsen = 4
    """Display various statistics during coarsening."""

    refine = 8
    """Display various statistics during refinement."""

    ipart = 16
    """Display various statistics during initial partitioning."""

    moveinfo = 32
    """Display detailed information about vertex moves during
    refinement."""

    sepinfo = 64
    """Display information about vertex separators."""

    conninfo = 128
    """Display information related to the minimization of
    subdomain connectivity."""

    contiginfo = 256
    """Display information related to the elimination of
    connected components."""


@enum.unique
class MetisRStatus(enum.IntEnum):
    """Return codes by METIS."""

    ok = 1
    """Returned normally."""

    error_input = -2
    """Returned due to erroneous inputs and/or options."""
    
    error_memory = -3
    """Returned due to insufficient memory."""
    
    error = -4
    """Some other errors."""



from nxmetis import enums
__all__ = ['MetisError']


class MetisError(Exception):
    """Error returned by METIS"""

    def __init__(self, rstatus, msg):
        """Initializes a MetisError object.

        Parameters
        ----------
        rstatus : int
            Error code returned by METIS.

        msg : string
            Error message returned by METIS.
        """
        super(MetisError, self).__init__('{0}: {1}'.format(
            enums.MetisRStatus(rstatus).name, msg))



import itertools
import nose.tools

import networkx as nx

import nxmetis
from nxmetis import exceptions
from nxmetis import metis
from nxmetis import types


def make_cycle(n):
    xadj = list(range(0, 2 * n + 1, 2))
    adjncy = list(
        itertools.chain.from_iterable(
            zip(itertools.chain([n - 1], range(n - 1)),
                itertools.chain(range(1, n), [0]))))
    return xadj, adjncy


class TestMetis(object):

    def setUp(self):
        self.node_list = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                          1, 2, 3, 4, 5, 6]
        self.G = nx.Graph()
        self.G.add_path(self.node_list)
        self.G.add_edge(self.node_list[-1], self.node_list[0])

    def test_node_nested_dissection_unweighted(self):
        node_ordering = nxmetis.node_nested_dissection(self.G)
        nose.tools.assert_equal(len(self.G), len(node_ordering))
        nose.tools.assert_equal(set(self.G), set(node_ordering))

        # Tests for exercising package's ability to handle self-loops
        # METIS crashes on self loops. networkx-metis should not
        self.G.add_edge(1, 1)
        self.G.add_edge('a', 'a')
        node_ordering = nxmetis.node_nested_dissection(self.G)
        nose.tools.assert_equal(len(self.G), len(node_ordering))
        nose.tools.assert_equal(set(self.G), set(node_ordering))

    def test_partition(self):
        partition = nxmetis.partition(self.G, 4)
        # When we choose one node from one part of the partitioned Graph,
        # It must be adjacent to one or more of the nodes in the same part.
        # This is to verify the continuity of the chain of nodes.
        parts = partition[1]  # List containing partitioned node lists

        nose.tools.assert_equal(partition[0], 4)
        nose.tools.assert_equal(len(partition[1]), 4)

        for part in parts:
            nose.tools.assert_not_equal(0, len(part))  # Non-empty set
            nose.tools.assert_equal(len(part), len(set(part)))  # Duplicate-free
            nose.tools.ok_(nx.is_connected(self.G.subgraph(part)))  # Connected

        # Disjoint sets
        for part1, part2 in itertools.combinations(parts, 2):
            nose.tools.assert_equal(set(), set(part1) & set(part2))

        # These parts must be exhaustive with the node list of the Graph
        parts_combined = parts[0] + parts[1] + parts[2] + parts[3]
        nose.tools.assert_equal(set(parts_combined), set(self.G))

    def test_vertex_separator(self):
        sep, part1, part2 = nxmetis.vertex_separator(self.G)

        # The two separator nodes must not be present in the
        # two bisected chains
        nose.tools.ok_(sep[0] not in part1)
        nose.tools.ok_(sep[0] not in part2)
        nose.tools.ok_(sep[1] not in part1)
        nose.tools.ok_(sep[1] not in part2)

        # There should be two different separator nodes
        nose.tools.assert_equal(len(sep), 2)
        nose.tools.assert_not_equal(sep[0], sep[1])

        # The lists should be exhaustive with the node list of the Graph
        nose.tools.assert_equal(set(sep) | set(part1) | set(part2),
                                set(self.G))

        # The parts must be disjoint sets
        nose.tools.assert_equal(set(), set(part1) & set(part2))

        # Non-empty set
        nose.tools.assert_not_equal(len(part1), 0)
        nose.tools.assert_not_equal(len(part2), 0)

        # Duplicate-free
        nose.tools.assert_equal(len(part1), len(set(part1)))
        nose.tools.assert_equal(len(part2), len(set(part2)))

        # Connected
        nose.tools.ok_(nx.is_connected(self.G.subgraph(part1)))
        nose.tools.ok_(nx.is_connected(self.G.subgraph(part2)))

    def test_MetisOptions(self):
        n = 16
        xadj, adjncy = make_cycle(n)
        options = types.MetisOptions(niter=-2)
        nose.tools.assert_raises_regexp(exceptions.MetisError,
                                        'Input Error: Incorrect niter.',
                                        metis.part_graph, xadj, adjncy, 2,
                                        options=options)



import numbers

from nxmetis import enums
from nxmetis import metis

__all__ = ['MetisOptions']


class MetisOptions(object):
    """Options controlling behaviors of METIS algorithms."""

    def __init__(self, **kwargs):
        """Initializes a MetisOptions object. Values can be provided
        for some parameters as arguments.

        Example
        -------
        >>> options = MetisOptions(ncuts=2, niter=100)

        """
        metis.set_default_options(self)
        for key, value in kwargs.items():
            if value is not None:
                setattr(self, key, value)

    @property
    def ptype(self):
        """Types of Partitioning method.

            ==== =================================
            rb   Multilevel recursive bisectioning
            kway Multilevel `k`-way partitioning
            ==== =================================
        """
        return self._ptype

    @ptype.setter
    def ptype(self, value):
        self._ptype = enums.MetisPType(value)

    @property
    def objtype(self):
        """Type of objective.

            === =======================================
            cut Edge-cut minimization
            vol Total communication volume minimization
            === =======================================
        """
        return self._objtype

    @objtype.setter
    def objtype(self, value):
        self._objtype = enums.MetisObjType(value)

    @property
    def ctype(self):
        """Matching scheme to be used during coarsening.

            ==== ==========================
            rm   Random matching
            shem Sorted heavy-edge matching
            ==== ==========================
        """
        return self._ctype

    @ctype.setter
    def ctype(self, value):
        self._ctype = enums.MetisCType(value)

    @property
    def iptype(self):
        """Algorithm used during initial partitioning.

            ====== ======================================================
            grow   Grow a bisection using a greedy strategy
            random Compute a bisection at random followed by a refinement
            edge   Derive a separator from an edge cut
            node   Grow a bisection using a greedy node-based strategy
            ====== ======================================================
       """
        return self._iptype

    @iptype.setter
    def iptype(self, value):
        self._iptype = enums.MetisIPType(value)

    @property
    def rtype(self):
        """Algorithm used for refinement.

            ========= ======================================
            fm        FM-based cut refinement
            greedy    Greedy-based cut and volume refinement
            sep2sided Two-sided node FM refinement
            sep1sided One-sided node FM refinement
            ========= ======================================
        """
        return self._rtype

    @rtype.setter
    def rtype(self, value):
        self._rtype = enums.MetisRType(value)

    @property
    def ncuts(self):
        """Number of cuts."""
        return self._ncuts

    @ncuts.setter
    def ncuts(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._ncuts = int(value)

    @property
    def nseps(self):
        """Number of separators."""
        return self._nseps

    @nseps.setter
    def nseps(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._nseps = int(value)

    @property
    def numbering(self):
        """Numbering scheme is used for the adjacency structure of a graph or the
        element-node structure of a mesh.

            ==== =================================
            zero C-style zero-based numbering
            one  Fortran-style one-based numbering
            ==== =================================
        """
        return self._numbering

    @numbering.setter
    def numbering(self, value):
        self._numbering = enums.MetisNumbering(value)

    @property
    def niter(self):
        """Number of refinement iterations."""
        return self._niter

    @niter.setter
    def niter(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._niter = value

    @property
    def seed(self):
        """Random number seed."""
        return self._seed

    @seed.setter
    def seed(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._seed = value

    @property
    def minconn(self):
        """Number of mimimum connectivity."""
        return self._minconn

    @minconn.setter
    def minconn(self, value):
        self._minconn = bool(value)

    @property
    def no2hop(self):
        """A boolean to perform a 2-hop matching."""
        return self._no2hop

    @no2hop.setter
    def no2hop(self, value):
        self._no2hop = bool(value)

    @property
    def contig(self):
        """A boolean to create contigous partitions."""
        return self._contig

    @contig.setter
    def contig(self, value):
        self._contig = bool(value)

    @property
    def compress(self):
        """A boolean to compress graph prior to ordering."""
        return self._compress

    @compress.setter
    def compress(self, value):
        self._compress = bool(value)

    @property
    def ccorder(self):
        """A boolean to detect & order connected components separately."""
        return self._ccorder

    @ccorder.setter
    def ccorder(self, value):
        self._ccorder = bool(value)

    @property
    def pfactor(self):
        """Prunning factor for high degree vertices."""
        return self._pfactor

    @pfactor.setter
    def pfactor(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._pfactor = value

    @property
    def ufactor(self):
        """User-supplied ufactor."""
        return self._ufactor

    @ufactor.setter
    def ufactor(self, value):
        if not isinstance(value, numbers.Number) or value != int(value):
            raise ValueError('{0} is not an int'.format(repr(value)))
        self._ufactor = value

    @property
    def dbglvl(self):
        """Amount of progress/debugging information will be printed during the
        execution of the algorithms. Can be combined by bit-wise OR.

            ========== ======================================================
            info       Print various diagnostic messages
            time       Perform timing analysis
            coarsen    Display various statistics during coarsening
            refine     Display various statistics during refinement
            ipart      Display various statistics during initial partitioning
            moveinfo   Display detailed information about vertex moves during
                       refinement
            sepinfo    Display information about vertex separators
            conninfo   Display information related to the minimization of
                       subdomain connectivity
            contiginfo Display information related to the elimination of
                       connected components
            ========== ======================================================
        """
        return self._dbglvl

    @dbglvl.setter
    def dbglvl(self, value):
        if (not isinstance(value, numbers.Number) or value != int(value) or
            not (value == -1 or (value >= 0 and value < 512) or
                 (value >= 2048 and value < 2560))):
            raise ValueError('{0} is not a valid dbglvl'.format(repr(value)))
        self._dbglvl = int(value)

    def __repr__(self):
        names = ['ptype', 'objtype', 'ctype', 'iptype', 'rtype', 'ncuts',
                 'nseps', 'numbering', 'niter', 'seed', 'minconn', 'no2hop',
                 'contig', 'compress', 'ccorder', 'pfactor', 'ufactor']
        return '{0}({1})'.format(
            self.__class__.__name__,
            ', '.join('{0}={1}'.format(name, repr(getattr(self, name)))
                      for name in names))
